
import numpy
from scipy.cluster.vq import vq, kmeans, whiten

"""
A spectral clustering method.
"""


class SpectralClusterer(object):
    def __init__(self, k, normLaplacian=False):
        """
        Initialise the object. 

        :param k: the number of clusters
        :type k: :class:`int`

        :param normLaplacian: whether to use the normalised Laplacian
        :type normLaplacian: :class:`bool`
        """
        self.k = k
        self.normLaplacian = normLaplacian 

    def cluster(self, graph):
        """
        Take a graph and cluster using spectral methods. The parameter k controls
        how many clusters to look for and normLaplacian specified whether to use
        the normalised Laplacian (as in Shi and Malik (2000)) or the unnormalised
        Laplacian.

        :param graph: the graph to cluster
        :type graph: :class:`apgl.graph.AbstractMatrixGraph`

        :returns:  An array of size graph.getNumVertices() of cluster membership 
        """
        #TODO: Use sparse eigenvalue computations for sparse matrix 
        if self.normLaplacian:
            L = graph.normalisedLaplacianRw()
        else:
            L = graph.laplacianMatrix()
            
        w, V = numpy.linalg.eig(L)

        sortedInds = numpy.argsort(w)
        sortedInds = sortedInds[0:self.k]
        V = V[:, sortedInds]

        whitened = whiten(V)
        codebook, distortion = kmeans(whitened, self.k)
        code, dist = vq(whitened, codebook)

        return code
