

import logging
import sys
import matplotlib
import numpy
import pickle
import os.path
matplotlib.use('WXAgg') # do this before importing pylab
import matplotlib.pyplot as plt
from datetime import date
from apgl.util.PathDefaults import PathDefaults
from apgl.util.DateUtils import DateUtils
from apgl.util.Latex import Latex 
from apgl.graph import * 
from apgl.viroscopy.HIVGraphReader import HIVGraphReader
from apgl.viroscopy.HIVGraphStatistics import HIVGraphStatistics

"""
This script computes some basic statistics on the growing graph. We currently
combine both infection and detection graphs (which are both directed) and hence
look at the contact graph. 
"""
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
numpy.set_printoptions(suppress=True)
numpy.set_printoptions(linewidth=60)

hivReader = HIVGraphReader()
graph = hivReader.readHIVGraph()
fInds = hivReader.getIndicatorFeatureIndices()

figureDir = PathDefaults.getOutputDir() + "viroscopy/figures/contact/"
resultsDir = PathDefaults.getOutputDir() + "viroscopy/"

#The set of edges indexed by zeros is the contact graph
#The ones indexed by 1 is the infection graph
edgeTypeIndex1 = 0
edgeTypeIndex2 = 1
sGraphContact = graph.getSparseGraph(edgeTypeIndex1)
sGraphInfect = graph.getSparseGraph(edgeTypeIndex2)
sGraphContact = sGraphContact.union(sGraphInfect)

sGraph = sGraphContact
#sGraph = sGraph.subgraph(range(0, 200))

graphStats = GraphStatistics()
statsArray = graphStats.scalarStatistics(sGraph, False)
slowStats = True
saveResults = False

logging.info(sGraph)
logging.info("Number of vertices: " + str(statsArray[graphStats.numVerticesIndex]))
logging.info("Number of features: " + str(sGraph.getVertexList().getNumFeatures()))
logging.info("Number of edges: " + str(statsArray[graphStats.numEdgesIndex]))
logging.info("Largest component is " + str(statsArray[graphStats.maxComponentSizeIndex]))
logging.info("Number of components " + str(statsArray[graphStats.numComponentsIndex]))

#sGraph = sGraph.subgraph(components[componentIndex])
vertexArray = sGraph.getVertexList().getVertices(range(0, sGraph.getVertexList().getNumVertices()))
logging.info("Size of graph we will use: " + str(sGraph.getNumVertices()))

#Some indices
dobIndex = fInds["birthDate"]
detectionIndex = fInds["detectDate"]
deathIndex = fInds["deathDate"]
genderIndex = fInds["gender"]
orientationIndex = fInds["orient"]
contactIndex = fInds["contactTrace"]
volIndex = fInds["volTest"]
doctorIndex = fInds["recommendVisit"]

ages = vertexArray[:, detectionIndex] - vertexArray[:, dobIndex]
deaths = vertexArray[:, deathIndex] - vertexArray[:, detectionIndex]
detections = vertexArray[:, detectionIndex]
meanAgeAtDetection = numpy.mean(ages)
stdAgeAtDetection = numpy.std(ages)
meanDeathTime = numpy.mean(deaths)
stdDeathTime = numpy.std(deaths)

startYear = 1900
daysInYear = 365
daysInMonth = 30 
logging.info("Mean age at detection: " + str(meanAgeAtDetection/daysInYear))
logging.info("Std age at detection: " + str(stdAgeAtDetection/daysInYear))

q = 0.9

#Properties of vertex values 
detectionAges = []
deathAfterInfectAges = []
deathAges = []
homoMeans = []
maleSums = []
femaleSums = []
heteroSums = []
biSums = []

contactSums = []
voluntarySums = []
recommendSums = []

monthStep = 3
dayList = range(int(numpy.min(detections)), int(numpy.max(detections)), daysInMonth*monthStep)
#Day 0 is the start
absDayList = [float(i-numpy.min(detections)) for i in dayList]
subgraphIndicesList = []

#Locations and labels for years
locs = range(0, int(absDayList[-1]), daysInYear*2)
labels = numpy.arange(1986, 2006, 2)

logging.info(dayList)
logging.info(absDayList)
plotInd = 1

#First, compute a series of properties based on vertex values
logging.info("Computing vertex stats")

for i in dayList:
    logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
    subgraphIndices = numpy.nonzero(detections <= i)[0]
    subgraphIndicesList.append(subgraphIndices)
    subgraph = sGraph.subgraph(subgraphIndices)

    #Values based on the properties of the vertices
    subgraphVertexArray = subgraph.getVertexList().getVertices(range(0, subgraph.getNumVertices()))
    detectionAges.append(numpy.mean((subgraphVertexArray[:, detectionIndex] - subgraphVertexArray[:, dobIndex]))/daysInYear)
    deathAfterInfectAges.append((numpy.mean(subgraphVertexArray[:, deathIndex] - subgraphVertexArray[:, detectionIndex]))/daysInYear)
    deathAges.append(numpy.mean((subgraphVertexArray[:, deathIndex] - subgraphVertexArray[:, dobIndex]))/daysInYear)
    homoMeans.append(numpy.mean(subgraphVertexArray[:, orientationIndex]))

    contactSums.append(numpy.sum(subgraphVertexArray[:, contactIndex]))
    voluntarySums.append(numpy.sum(subgraphVertexArray[:, volIndex]))
    recommendSums.append(numpy.sum(subgraphVertexArray[:, doctorIndex]))

    heteroSums.append(numpy.sum(subgraphVertexArray[:, orientationIndex]==0))
    biSums.append(numpy.sum(subgraphVertexArray[:, orientationIndex]==1))

    femaleSums.append(numpy.sum(subgraphVertexArray[:, genderIndex]==1))
    maleSums.append(numpy.sum(subgraphVertexArray[:, genderIndex]==0))

def plotVertexStats():
    global plotInd 

    plt.figure(plotInd)
    plt.plot(absDayList, detectionAges)
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detection Age (years)")
    plt.savefig(figureDir + "DetectionMeansGrowth.eps")
    plotInd += 1

    plt.figure(plotInd)
    plt.plot(absDayList, heteroSums, 'r', absDayList, biSums, 'b')
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Heterosexual", "Bisexual"), loc="upper left")
    plt.savefig(figureDir + "OrientationGrowth.eps")
    plotInd += 1

    plt.figure(plotInd)
    plt.plot(absDayList, femaleSums, 'r', absDayList, maleSums, 'b')
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Female", "Male"), loc="upper left")
    plt.savefig(figureDir + "GenderGrowth.eps")
    plotInd += 1

    #Plot of detection types
    plt.figure(plotInd)
    plt.plot(absDayList, contactSums, 'r', absDayList, voluntarySums, 'b', absDayList, recommendSums, 'g')
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Contact Tracing", "Voluntary Visit", "Doctor Recommendation"), loc="upper left")
    plt.savefig(figureDir + "DetectionGrowth.eps")
    plotInd += 1

def plotScalarStats():
    logging.info("Computing scalar stats")

    resultsFileName = resultsDir + "ContactGrowthScalarStats.pkl"

    if saveResults:
        statsArray = graphStats.sequenceScalarStats(sGraph, subgraphIndicesList, slowStats)
        if os.path.isfile(resultsFileName):
            raise IOError("File exists: " + resultsFileName)
        output = open(resultsFileName, 'wb')
        pickle.dump(statsArray, output)
        output.close()
        logging.info("Saved results in to file " + resultsFileName)
    else:
        input = open(resultsFileName, 'rb')
        statsArray = pickle.load(input)
        logging.info("Loaded results from file " + resultsFileName)
    
        global plotInd

        #Output all the results into plots
        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.maxComponentSizeIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component size")
        plt.savefig(figureDir + "MaxComponentSizeGrowth.eps")
        plotInd += 1

        print(absDayList)
        print(statsArray[:, graphStats.maxComponentSizeIndex])

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.numComponentsIndex], 'r', absDayList, statsArray[:, graphStats.numNonSingletonComponentsIndex], 'b')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Number of components")
        plt.legend(("Size > 0", "Size > 1"), loc="upper left")
        plt.savefig(figureDir + "NumComponentsGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.meanComponentSizeIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Mean component size")
        plt.savefig(figureDir + "MeanComponentSizeGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.diameterIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component diameter")
        plt.savefig(figureDir + "MaxComponentDiameterGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.effectiveDiameterIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component effective diameter")
        plt.savefig(figureDir + "MaxComponentEffDiameterGrowth.eps")
        plotInd += 1

        #Make sure we don't include 0 in the array
        vertexIndex = numpy.argmax(statsArray[:, graphStats.numVerticesIndex] > 0)
        edgeIndex = numpy.argmax(statsArray[:, graphStats.numEdgesIndex] > 0)
        minIndex = numpy.maximum(vertexIndex, edgeIndex)

        plt.figure(plotInd)
        plt.plot(numpy.log(statsArray[minIndex:, graphStats.numVerticesIndex]), numpy.log(statsArray[minIndex:, graphStats.numEdgesIndex]))
        plt.xlabel("log(|V|)")
        plt.ylabel("log(|E|)")
        plt.savefig(figureDir + "LogVerticesEdgesGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.meanDegreeIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Mean degree")
        plt.savefig(figureDir + "MeanDegrees.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.densityIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Density")
        plt.savefig(figureDir + "DensityGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.powerLawIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Alpha")
        plt.savefig(figureDir + "PowerLawGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.geodesicDistanceIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Geodesic distance")
        plt.savefig(figureDir + "GeodesicGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.harmonicGeoDistanceIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Mean harmonic geodesic distance")
        plt.savefig(figureDir + "HarmonicGeodesicGrowth.eps")
        plotInd += 1

        #print(statsArray[:, graphStats.harmonicGeoDistanceIndex])

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.geodesicDistMaxCompIndex])
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component geodesic Distance")
        plt.savefig(figureDir + "MaxCompGeodesicGrowth.eps")
        plotInd += 1

        #Print out the mean geodesic distances for Hector
        
        #print(statsArray[:, graphStats.geodesicDistMaxCompIndex][1:])


def plotVectorStats():
    #Finally, compute some vector stats at various points in the graph
    logging.info("Computing vector stats")
    dayList2 = [DateUtils.getDayDelta(date(1989, 12, 31), startYear)]
    dayList2.append(DateUtils.getDayDelta(date(1993, 12, 31), startYear))
    dayList2.append(DateUtils.getDayDelta(date(1997, 12, 31), startYear))
    dayList2.append(DateUtils.getDayDelta(date(2001, 12, 31), startYear))
    dayList2.append(int(numpy.max(detections)))
    
    subgraphIndicesList = []
    global plotInd
    
    for i in dayList2:
        logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
        subgraphIndices = numpy.nonzero(detections <= i)[0]
        subgraphIndicesList.append(subgraphIndices)

    resultsFileName = resultsDir + "ContactGrowthVectorStats.pkl"

    if saveResults:
        statsDictList = graphStats.sequenceVectorStats(sGraph, subgraphIndicesList)
        if os.path.isfile(resultsFileName):
            raise IOError("File exists: " + resultsFileName)
        output = open(resultsFileName, 'wb')
        pickle.dump(statsDictList, output)
        output.close()
        logging.info("Saved results in to file " + resultsFileName)
    else:
        input = open(resultsFileName, 'rb')
        statsDictList = pickle.load(input)
        logging.info("Loaded results from file " + resultsFileName)

        triangleDistArray = numpy.zeros((len(dayList2), 100))
        componentsDistArray = numpy.zeros((len(dayList2), 3000))
        numVerticesEdgesArray = numpy.zeros((len(dayList2), 2), numpy.int)
        numVerticesEdgesArray[:, 0] = [len(sgl) for sgl in subgraphIndicesList]
        numVerticesEdgesArray[:, 1] = [sGraph.subgraph(sgl).getNumEdges() for sgl in subgraphIndicesList]

        binWidths = numpy.arange(0, 0.50, 0.05)
        eigVectorDists = numpy.zeros((len(dayList2), binWidths.shape[0]-1), numpy.int)

        #Note that death has a lot of missing values
        for j in range(len(dayList2)):
            dateStr = (str(DateUtils.getDateStrFromDay(dayList2[j], startYear)))
            logging.info(dateStr)
            statsDict = statsDictList[j]

            degreeDist = statsDict["outDegreeDist"]
            degreeDist = degreeDist/float(numpy.sum(degreeDist))

            eigenDist = statsDict["eigenDist"]
            eigenDist = numpy.log(eigenDist[eigenDist>=10**-1])

            hopCount = statsDict["hopCount"]
            hopCount = numpy.log10(hopCount)
            #hopCount = numpy.array(hopCount, numpy.float64)/numpy.max(hopCount)

            triangleDist = statsDict["triangleDist"]
            triangleDist = numpy.array(triangleDist, numpy.float64)/numpy.sum(triangleDist)
            triangleDistArray[j, 0:triangleDist.shape[0]] = triangleDist

            maxEigVector = statsDict["maxEigVector"]
            maxEigVector = numpy.flipud(numpy.sort(numpy.abs(maxEigVector)))
            #print(maxEigVector)
            eigVectorDists[j, :] = numpy.histogram(maxEigVector, binWidths)[0]

            componentsDist = statsDict["componentsDist"]
            componentsDist = numpy.array(componentsDist, numpy.float64)/numpy.sum(componentsDist)
            componentsDistArray[j, 0:componentsDist.shape[0]] = componentsDist

            plotInd2 = plotInd

            plt.figure(plotInd2)
            plt.plot(numpy.arange(degreeDist.shape[0]), degreeDist, label=dateStr)
            plt.xlabel("Degree")
            plt.ylabel("Probability")
            plt.ylim((0, 0.5))
            plt.savefig(figureDir + "DegreeDist" +  ".eps")
            plt.legend()
            plotInd2 += 1

            """
            plt.figure(plotInd2)
            plt.plot(numpy.arange(eigenDist.shape[0]), eigenDist, label=dateStr)
            plt.xlabel("Eigenvalue rank")
            plt.ylabel("log(Eigenvalue)")
            plt.savefig(figureDir + "EigenDist" +  ".eps")
            plt.legend()
            plotInd2 += 1
            """

            #How does kleinberg do the hop plots 
            plt.figure(plotInd2)
            plt.plot(numpy.arange(hopCount.shape[0]), hopCount, label=dateStr)
            plt.xlabel("k")
            plt.ylabel("log10(pairs)")
            plt.ylim( (2.5, 7) )
            plt.legend(loc="lower right")
            plt.savefig(figureDir + "HopCount" + ".eps")
            plotInd2 += 1

            
            plt.figure(plotInd2)
            plt.plot(numpy.arange(maxEigVector.shape[0]), maxEigVector, label=dateStr)
            plt.xlabel("Rank")
            plt.ylabel("log(eigenvector coefficient)")
            plt.savefig(figureDir + "MaxEigVector" +  ".eps")
            plt.legend()
            plotInd2 += 1
            

        precision = 4
        dateStrList = [DateUtils.getDateStrFromDay(day, startYear) for day in dayList2]

        print("Edges and vertcies")
        print(Latex.listToRow(dateStrList))
        print(Latex.array2DToRows(numVerticesEdgesArray.T, precision))

        print("\n\nEigenvector distribution")
        print(Latex.array1DToRow(binWidths[1:]) + "\\\\")
        print(Latex.array2DToRows(eigVectorDists))

        print("\n\nDistribution of component sizes")
        componentsDistArray = componentsDistArray[:, 0:componentsDist.shape[0]]
        nonZeroCols = numpy.sum(componentsDistArray, 0)!=0
        componentsDistArray = numpy.r_[numpy.array([numpy.arange(componentsDistArray.shape[1])[nonZeroCols]]), componentsDistArray[:, nonZeroCols]]
        print(Latex.listToRow(dateStrList))
        print(Latex.array2DToRows(componentsDistArray.T, precision))

        print("\n\nDistribution of triangle participations")
        triangleDistArray = triangleDistArray[:, 0:triangleDist.shape[0]]
        nonZeroCols = numpy.sum(triangleDistArray, 0)!=0
        triangleDistArray = numpy.r_[numpy.array([numpy.arange(triangleDistArray.shape[1])[nonZeroCols]])/2, triangleDistArray[:, nonZeroCols]]
        print(Latex.listToRow(dateStrList))
        print(Latex.array2DToRows(triangleDistArray. T, precision))

def plotOtherStats():
    #Let's look at geodesic distances in subgraphs and communities
    logging.info("Computing other stats")

    resultsFileName = resultsDir + "ContactGrowthOtherStats.pkl"
    hivGraphStats = HIVGraphStatistics(fInds)

    if saveResults:
        statsArray = hivGraphStats.sequenceScalarStats(sGraph, subgraphIndicesList)

        if os.path.isfile(resultsFileName):
            raise IOError("File exists: " + resultsFileName)
        
        output = open(resultsFileName, 'wb')
        pickle.dump(statsArray, output)
        output.close()
        logging.info("Saved results in to file " + resultsFileName)
    else:
        input = open(resultsFileName, 'rb')
        statsArray = pickle.load(input)
        logging.info("Loaded results from file " + resultsFileName)

        global plotInd

        msmGeodesic = statsArray[:, hivGraphStats.msmGeodesicIndex]
        msmGeodesic[msmGeodesic < 0] = 0
        msmGeodesic[msmGeodesic == float('inf')] = 0

        #Output all the results into plots
        plt.figure(plotInd)
        plt.plot(absDayList, msmGeodesic, 'r', absDayList, statsArray[:, hivGraphStats.mostConnectedGeodesicIndex], 'b')
        plt.xticks(locs, labels)
        #plt.ylim([0, 0.1])
        plt.xlabel("Year")
        plt.ylabel("Mean harmonic geodesic distance")
        plt.legend(("MSM individuals", "Top 10% degree"), loc="upper right")
        plt.savefig(figureDir + "MSM10Geodesic" + ".eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, hivGraphStats.menSubgraphGeodesicIndex])
        plt.xticks(locs, labels)
        plt.ylim([0, 100.0])
        plt.xlabel("Year")
        plt.ylabel("Mean harmonic geodesic distance")
        plt.savefig(figureDir + "MenSubgraphGeodesic" + ".eps")
        plotInd += 1


def clusterVertices():
    """
    Cluster the vertices in the largest component of the growing graph and save the results
    """
    logging.info("Computing clustering")

    resultsFileName = resultsDir + "ContactGrowthClustering.pkl"

    if saveResults:
        statsArray = graphStats.sequenceScalarStats(sGraph, subgraphIndicesList, slowStats)
        if os.path.isfile(resultsFileName):
            raise IOError("File exists: " + resultsFileName)
        output = open(resultsFileName, 'wb')
        pickle.dump(statsArray, output)
        output.close()
        logging.info("Saved results in to file " + resultsFileName)
    else:
        pass 

#plotVertexStats()
#plotScalarStats()
#plotVectorStats()
plotOtherStats()
plt.show()

"""
Also clustering (4 clusters)
Probability of adding node based on degree - try to find how we can generate data.
Mean Time between first and last infection for each person
"""