import logging
import numpy
from apgl.graph import *
from apgl.util import *
from apgl.viroscopy.model.HIVGraph import HIVGraph
from apgl.viroscopy.model.HIVVertices import HIVVertices

"""
This class models an epidemic occuring via sexual contact. 
"""

class HIVEpidemicModel():
    def __init__(self, graph, rates):
        """
        Initialise with a graph of HIVIndividuals.
        """
        Parameter.checkClass(graph, HIVGraph)

        self.graph = graph
        self.rates = rates 

    def computeContactDetectRates(self, infectedList, removedSet, t):
        #This is the slowest point the code
        contactRates = self.rates.contactRates(infectedList, t)
        contactTracingRates = self.rates.contactTracingRates(infectedList, removedSet, t)
        randomDetectRates = self.rates.randomDetectionRates(infectedList, t)

        return contactRates, randomDetectRates, contactTracingRates

    def simulate(self, T, recordStep=5):
        """
        Simulate epidemic propogation until there are no more infectives or
        time T is reached. 
        """
        Parameter.checkFloat(T, 0.0, float('inf'))
        Parameter.checkInt(recordStep, 0, float('inf'))

        susceptibleSet = self.graph.getSusceptibleSet()
        infectedSet = self.graph.getInfectedSet()
        removedSet = self.graph.getRemovedSet()

        infectedList = list(infectedSet)
        removedList = list(removedSet)

        vList = self.graph.getVertexList()
        t = 0
        times = []
        #A list of lists of infected indices 
        infectedIndices = []
        removedIndices = []
        nextStep = 0

        #Now, start the simulation
        while t < T and len(infectedSet) != 0:
            logging.info("t=" + str(t) + " S=" + str(len(susceptibleSet)) + " I=" + str(len(infectedSet)) + " R=" + str(len(removedSet)))
            contactRates, randomDetectRates, contactTracingRates = self.computeContactDetectRates(infectedList, removedSet, t)

            sigmat = contactRates.sum()
            muRSt = numpy.sum(randomDetectRates)
            muCTt = numpy.sum(contactTracingRates)
            rhot = sigmat + muRSt + muCTt

            #Now generate random variable which is the advancement in time
            tauPrime = numpy.random.exponential(1/rhot)
            t = t + tauPrime

            #Now compute the probabilities of each event type
            contactProb = sigmat/rhot
            detectionRandom = muRSt/rhot
            detectionContact = muCTt/rhot 

            #Compute random variable
            p = numpy.random.rand()

            if p < contactProb:
                contactIndices = contactRates.nonzero()
                eventInd = Util.randomChoice(contactRates[contactIndices].toarray().ravel())[0]
                infectedIndex = infectedList[contactIndices[0][eventInd]]
                contactIndex = contactIndices[1][eventInd]
                #Note that each time a sexual contact occurs we weight the edge with the time 
                self.rates.contactEvent(infectedIndex, contactIndex, t)
                #logging.info("Sexual contact between vertices " + str(infectedIndex) + " and " + str(contactIndex))

                #Check if the contact results in an infection
                q = numpy.random.rand()
                if q < self.rates.infectionProbability(infectedIndex, contactIndex, t):
                    vList.setInfected(contactIndex, t)
                    infectedSet.add(contactIndex)
                    susceptibleSet.remove(contactIndex)
                    #logging.info("Infection between vertices " + str(infectedIndex) + " and " + str(contactIndex))
            elif p >= contactProb and p < contactProb+detectionRandom:
                eventInd = Util.randomChoice(randomDetectRates)
                newDetectedIndex = infectedList[eventInd]
                vList.setDetected(newDetectedIndex, t, HIVVertices.randomDetect)
                removedSet.add(newDetectedIndex)
                infectedSet.remove(newDetectedIndex)
                #logging.info("Random detection made on vertex " + str(newDetectedIndex))
            else:
                eventInd = Util.randomChoice(contactTracingRates)
                newDetectedIndex = infectedList[eventInd]
                vList.setDetected(newDetectedIndex, t, HIVVertices.contactTrace)
                removedSet.add(newDetectedIndex)
                infectedSet.remove(newDetectedIndex)
                #logging.info("Contact tracing detection made on vertex " + str(newDetectedIndex))

            infectedList = list(infectedSet)
            removedList = list(removedSet)

            if t >= nextStep:
                infectedIndices.append(infectedList)
                removedIndices.append(removedList)
                times.append(t)
                nextStep += recordStep 

        return times, infectedIndices, removedIndices, self.graph