import pickle
import numpy
import logging
import sys 
import matplotlib.pyplot as plt

from apgl.graph import *
from apgl.util.PathDefaults import PathDefaults
from apgl.viroscopy.model.HIVIndividual import HIVIndividual

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

outputDir = PathDefaults.getOutputDir() + "viroscopy/"
graphFileName = outputDir + "epidemicGraph.pkl"
indicesFileName = outputDir + "epidemicDetectIndices.pkl"
timesFileName = outputDir + "epidemicTimes.pkl"

graphFile = open(graphFileName, 'rb')
graph = pickle.load(graphFile)
graphFile.close()
logging.info("Loaded a graph: " + str(graph))

W = graph.getSparseWeightMatrix().tocsr()
graph.setWeightMatrixSparse(W)
logging.info("Converted graph to type " + str(graph.weightMatrixType()))

indicesFile = open(indicesFileName, 'rb')
removedIndices = pickle.load(indicesFile)
indicesFile.close()
logging.info("Loaded " + str(len(removedIndices)) + " sets of indices")

timesFile = open(timesFileName, 'rb')
times = pickle.load(timesFile)
timesFile.close()
logging.info("Loaded times")

graphStats = GraphStatistics()
statsArray = graphStats.sequenceScalarStats(graph, removedIndices, True)

finalRemovedInds = removedIndices[-1]
removedGraph = graph.subgraph(finalRemovedInds)
statsDict = graphStats.vectorStatistics(removedGraph)


#Let's plot the statistics
plotInd = 1
plt.figure(plotInd)
plt.plot(times, statsArray[:, graphStats.maxComponentSizeIndex])
plt.xlabel("Time (days)")
plt.ylabel("Max Component Size")
plotInd += 1

plt.figure(plotInd)
plt.plot(times, statsArray[:, graphStats.numComponentsIndex])
plt.xlabel("Time (days)")
plt.ylabel("Number of components")
plotInd += 1

plt.figure(plotInd)
plt.plot(times, statsArray[:, graphStats.meanComponentSizeIndex])
plt.xlabel("Time (days)")
plt.ylabel("Mean Component Sizes")
plotInd += 1

plt.figure(plotInd)
plt.plot(times, statsArray[:, graphStats.effectiveDiameterIndex])
plt.xlabel("Time (days)")
plt.ylabel("Max Component Effective Diameter")
plotInd += 1

degreeDist = statsDict["outDegreeDist"]

plt.figure(plotInd)
plt.plot(numpy.arange(degreeDist.shape[0]), degreeDist/float(numpy.sum(degreeDist)))
plt.xlabel("Degree")
plt.ylabel("Probability")
plotInd += 1


#Make sure we don't include 0 in the array
vertexIndex = numpy.argmax(statsArray[:, graphStats.numVerticesIndex] > 0)
edgeIndex = numpy.argmax(statsArray[:, graphStats.numEdgesIndex] > 0)
minIndex = numpy.maximum(vertexIndex, edgeIndex)

plt.figure(plotInd)
plt.plot(numpy.log(statsArray[minIndex:, graphStats.numVerticesIndex]), numpy.log(statsArray[minIndex:, graphStats.numEdgesIndex]))
plt.xlabel("log(numVertices)")
plt.ylabel("log(numEdges)")
plotInd += 1


plt.figure(plotInd)
plt.plot(times, statsArray[:, graphStats.meanDegreeIndex])
plt.xlabel("Time (days)")
plt.ylabel("meanDegrees")
plotInd += 1

#Plot also some properties of the vertices (include detection types)
numMales = numpy.zeros((len(removedIndices)))
numFemales = numpy.zeros((len(removedIndices)))
numHetros = numpy.zeros((len(removedIndices)))
numBis = numpy.zeros((len(removedIndices)))
numRandomDetects = numpy.zeros((len(removedIndices)))
numContactDetects = numpy.zeros((len(removedIndices)))

for i in range(len(removedIndices)):
    currentSubgraph = graph.subgraph(removedIndices[i])

    for j in range(currentSubgraph.getNumVertices()):
        if currentSubgraph.getVertex(j).getGender() == HIVIndividual.male:
            numMales[i] += 1
        else:
            numFemales[i] += 1

        if currentSubgraph.getVertex(j).getOrientation() == HIVIndividual.bi:
            numBis[i] += 1
        else:
            numHetros[i] += 1

        if currentSubgraph.getVertex(j).getDetectionType() == HIVIndividual.randomDetect:
            numRandomDetects[i] += 1
        else:
            numContactDetects[i] += 1

plt.figure(plotInd)
plt.plot(times, numMales, 'r', times, numFemales, 'b')
plt.xlabel("Time (days)")
plt.ylabel("Detections")
plt.legend(("Male", "Female"), loc="upper left")
plotInd += 1

plt.figure(plotInd)
plt.plot(times, numHetros, 'r', times, numBis, 'b')
plt.xlabel("Time (days)")
plt.ylabel("Detections")
plt.legend(("Heterosexual", "Bisexual"), loc="upper left")
plotInd += 1

plt.figure(plotInd)
plt.plot(times, numRandomDetects, 'r', times, numContactDetects, 'b')
plt.xlabel("Time (days)")
plt.ylabel("Detections")
plt.legend(("Random Detection", "Contact Tracing"), loc="upper left")
plotInd += 1

plt.show()
