
import numpy
import scipy.sparse

from apgl.viroscopy.model.HIVVertices import HIVVertices
from apgl.util.Util import *

"""
Model the contact rate of an infected individual and other susceptibles.
"""
class HIVRates():
    def __init__(self, graph):
        self.graph = graph
        self.degreeSequence = graph.outDegreeSequence()
        self.V = self.graph.getVertexList().getVertices(range(0,self.graph.getNumVertices()))

        #Parameters for sexual contact
        self.alpha = 2.0
        self.newContactChance = 5.0
        self.heteroContactRate = 0.1
        self.biContactRate = 0.5

        #Infection probabilities 
        self.womanManInfectProb = 0.3
        self.manWomanHeteroInfectProb = 0.2
        self.manBiInfectProb = 0.7
        self.defaultInfectProb = 0.5

        #Random detection
        self.randDetectRate = 11.0

        #Contact tracing parameters 
        self.ctRatePerPerson = 0.3
        #The start and end time of contact tracing
        self.ctStartTime = 180
        self.ctEndTime = 730

        #contactTimesArr is an array of the index of the last sexual contact and contact time
        self.contactTimesArr = numpy.zeros((graph.getNumVertices(), 2))
        self.neighboursList = [] 

        for i in range(graph.getNumVertices()):
            self.neighboursList.append(graph.neighbours(i))
            self.contactTimesArr[i, :] = self.contactTimes(i)

    def contactTimes(self, vertexInd):
        """
        Returns an array with two elements, the first is the index of the last
        sexual contact (-1 if none), and the 2nd is the time of contact. 
        """
        neighbours = self.neighboursList[vertexInd]
        edges = numpy.c_[numpy.ones(neighbours.shape[0], numpy.int32)*vertexInd, neighbours]
        edgeValues = self.graph.getEdgeValues(edges)

        contactTimes = numpy.array([-1, -1])

        if edgeValues != []:
            ind = numpy.argmax(edgeValues)
            contactTimes = numpy.array([neighbours[ind], edgeValues[ind]])

        return contactTimes

    def contactEvent(self, vertexInd1, vertexInd2, t):
        """
        Indicates a sexual contact event between two vertices. 
        """
        if self.graph.getEdge(vertexInd1, vertexInd2) == None:
            self.degreeSequence[vertexInd1] += 1
            self.degreeSequence[vertexInd2] += 1

        self.graph.addEdge(vertexInd1, vertexInd2, t)
        self.neighboursList[vertexInd1] = self.graph.neighbours(vertexInd1)
        self.neighboursList[vertexInd2] = self.graph.neighbours(vertexInd2)
        self.contactTimesArr[vertexInd1, :] = self.contactTimes(vertexInd1)
        self.contactTimesArr[vertexInd2, :] = self.contactTimes(vertexInd2)
            
    def contactRates(self, infectedList, t):
        """
        Work out contact rates between all infected and all others. What is
        computational complexity?
        """
        if len(infectedList) == 0:
            return numpy.array([])

        contactRates = scipy.sparse.lil_matrix((len(infectedList), self.graph.getNumVertices()))
        
        #First choose a set of new contacts based on the degree sequence
        #These contacts can be any one of the other contacts
        possibleContacts = numpy.zeros((len(infectedList), 2), numpy.int)
        possibleContacts[:, 0] = self.contactTimesArr[infectedList, 0]
        possibleContacts[:, 1] = Util.randomChoice(self.degreeSequence+1, len(infectedList))

        #If someone has no previous contact, make sure the probability is zero
        #Could exclude zero probability evens from randomChoice if that speed things up 
        possibleContactTimes = numpy.zeros((len(infectedList), 2))
        possibleContactTimes[:, 0] = self.contactTimesArr[infectedList, 1]
        possibleContactTimes[:, 0][possibleContactTimes[:, 0]==-1] = float('inf')
        possibleContactTimes[:, 1] = numpy.ones(len(infectedList))*self.newContactChance

        possibleContactProbs = possibleContactTimes**-self.alpha

        contactInds = Util.randomChoice(possibleContactProbs).ravel()
        contacts = possibleContacts[(numpy.arange(possibleContacts.shape[0]), contactInds)]

        #We compute sexual contact rates between infected and everyone else
        for i in xrange(len(infectedList)):
            vertexInd1 = infectedList[i]
            vertex1 = self.V[vertexInd1, :]

            vertexInd2 = contacts[i]
            vertex2 = self.V[vertexInd2, :]

            if vertex1[HIVVertices.genderIndex] != vertex2[HIVVertices.genderIndex]:
                contactRates[i, vertexInd2] = self.heteroContactRate
            elif vertex1[HIVVertices.genderIndex] == vertex2[HIVVertices.genderIndex] and vertex1[HIVVertices.orientationIndex]==HIVVertices.bi and vertex2[HIVVertices.orientationIndex]==HIVVertices.bi:
                contactRates[i, vertexInd2] = self.biContactRate
 
        return contactRates

    """
    Compute the infection probability between an infected and susceptible.
    """
    def infectionProbability(self, vertexInd1, vertexInd2, t):
        vertex1 = self.graph.getVertex(vertexInd1)
        vertex2 = self.graph.getVertex(vertexInd2)

        if vertex2[HIVVertices.stateIndex]==HIVVertices.infected or vertex2[HIVVertices.stateIndex]==HIVVertices.removed:
            return 0.0 

        #Gender 0 is female, 1 is male
        if vertex1[HIVVertices.genderIndex] == HIVVertices.female and vertex2[HIVVertices.genderIndex] == HIVVertices.male:
            return self.womanManInfectProb
        if vertex1[HIVVertices.genderIndex] == HIVVertices.male and vertex2[HIVVertices.genderIndex] == HIVVertices.female and vertex1[HIVVertices.orientationIndex]==HIVVertices.hetero:
            return self.manWomanHeteroInfectProb
        if vertex1[HIVVertices.genderIndex] == HIVVertices.male and vertex2[HIVVertices.genderIndex]==HIVVertices.bi:
            return self.manBiInfectProb
        else:
            return self.defaultInfectProb

    """
    Compute the detection rate of an infected which depends on the entire population.
    """
    def randomDetectionRates(self, infectedList, t):
        detectionRates = numpy.ones(len(infectedList))*self.randDetectRate/self.graph.getNumVertices()

        return detectionRates

    def contactTracingRates(self, infectedList, removedSet, t):
        """
        Compute the contact tracing detection rate of a list of infected individuals.
        """
        ctRates = numpy.zeros(len(infectedList))
        #A detected neigbour that happened at ctStartDate or earlier can result in detection
        ctStartDate = t - self.ctStartTime
        #A detected neigbour that happened later than cdEndDate can result in detection
        cdEndDate = t - self.ctEndTime

        removeIndices = self.V[:, HIVVertices.stateIndex]==HIVVertices.removed
        detectionTimes = self.V[:, HIVVertices.detectionTimeIndex]

        for i in range(len(infectedList)):
            vertexInd = infectedList[i]
            #We can find the set of detected neigbours here
            neighbours = self.neighboursList[vertexInd]

            #The rate is a sum of rates of detected contacts
            for ind in neighbours:
                if removeIndices[ind] and detectionTimes[ind] >= cdEndDate and detectionTimes[ind] <= ctStartDate:
                    ctRates[i] += self.ctRatePerPerson

        return ctRates