
import numpy 
import unittest
from apgl.viroscopy.model.HIVRates import HIVRates
from apgl.viroscopy.model.HIVGraph import HIVGraph
from apgl.viroscopy.model.HIVVertices import HIVVertices

class  HIVRateFuncsTestCase(unittest.TestCase):
    def setUp(self):
        numpy.set_printoptions(suppress=True)
        numpy.random.seed(21)

    def testContactEvent(self):
        undirected = True
        numVertices = 10
        graph = HIVGraph(numVertices, undirected)

        for i in range(numVertices):
            print(graph.getVertex(i))

        t = 0.2

        rates = HIVRates(graph)

        rates.contactEvent(0, 3, 0.2)
        rates.contactEvent(0, 9, 0.1)

        self.assertTrue((rates.degreeSequence == numpy.array([2,0,0,1,0,0,0,0,0, 1])).all())
        self.assertEquals(graph.getEdge(0, 3), 0.2)
        self.assertEquals(graph.getEdge(0, 9), 0.1)

        self.assertTrue((rates.contactTimesArr[0, :] == numpy.array([3, 0.2])).all())

    def testContactRates(self):
        undirected = True 
        numVertices = 10
        graph = HIVGraph(numVertices, undirected)


        vertex1Ind = 0
        susceptibleList = range(1, numVertices)
        t = 0.2

        rates = HIVRates(graph)
        contactRates = rates.contactRates([0, 5, 7], t)
        self.assertEquals(contactRates.shape[0], 3)
        self.assertEquals(contactRates.shape[1], 10)


        #Now we have that 0 had contact with another
        #graph.addEdge(0, 3, 0.1)
        #graph.addEdge(0, 9, 0.1)
        rates.contactEvent(0, 3, 0.2)
        rates.contactEvent(0, 9, 0.1)

        contactRates = rates.contactRates([0, 5, 7], t)
        infectIndices = contactRates.nonzero()[0]
        infectIndices = numpy.sort(infectIndices)

        #Test for duplicates in contacts 
        last = -1
        for i in range(infectIndices.shape[0]):
            if last == infectIndices[i]:
                self.fail("Infected indices should only have 1 contact")
            last = infectIndices[i]


    def testContactTracingRate(self):
        undirected = True
        numVertices = 10
        graph = HIVGraph(numVertices, undirected)

        t = 0.1

        graph.getVertexList().setInfected(0, t)

        rates = HIVRates(graph)
        rates.contactEvent(0, 3, 0.2)
        rates.contactEvent(0, 9, 0.1)

        t = 0.3
        graph.getVertexList().setInfected(3, t)
        graph.getVertexList().setInfected(9, t)

        t = 0.4
        graph.getVertexList().setDetected(0, t, HIVVertices.randomDetect)

        removedSet = graph.getRemovedSet()
        infectedList = [3]
        ctRates = rates.contactTracingRates(infectedList, removedSet, t)
        self.assertTrue((ctRates==numpy.array([0.0])).all())
        ctRates = rates.contactTracingRates(infectedList, removedSet, t+rates.ctStartTime)
        self.assertTrue((ctRates == numpy.array([rates.ctRatePerPerson])).all())

        rates.contactEvent(3, 5, t)
        graph.getVertexList().setInfected(5, t)
        graph.getVertexList().setDetected(5, t, HIVVertices.randomDetect)
        removedSet = graph.getRemovedSet()
        ctRates = rates.contactTracingRates(infectedList, removedSet, t+rates.ctStartTime)
        self.assertTrue((ctRates == numpy.array([rates.ctRatePerPerson*2])).all())
        
        rates.contactEvent(3, 6, t)
        graph.getVertexList().setInfected(6, t)
        removedSet = graph.getRemovedSet()
        ctRates = rates.contactTracingRates(infectedList, removedSet, t+rates.ctStartTime)
        self.assertTrue((ctRates == numpy.array([rates.ctRatePerPerson*2])).all())

        #Test contact tracing is within correct time period
        ctRates = rates.contactTracingRates(infectedList, removedSet, t+rates.ctEndTime)
        self.assertTrue((ctRates == numpy.array([rates.ctRatePerPerson*2])).all())

        ctRates = rates.contactTracingRates(infectedList, removedSet, t+rates.ctEndTime+1)
        self.assertTrue((ctRates == numpy.array([0.0])).all())

    def testRandomDetectionRates(self):
        undirected = True
        numVertices = 10
        graph = HIVGraph(numVertices, undirected)

        t = 0.1
        graph.getVertexList().setInfected(0, t)

        rates = HIVRates(graph)
        infectedList = [0, 2, 9]

        rdRates = rates.randomDetectionRates(infectedList, t)

        self.assertTrue((rdRates == numpy.ones(len(infectedList))*rates.randDetectRate/numVertices).all())

    def testInfectionProbability(self):
        #TODO
        pass 

if __name__ == '__main__':
    unittest.main()

